''*******************************************************************
''*       Title: VGA display 40x15 (single cog, ROM font, palette)  *
''*    FileName: VGA_40x15_display.spin                             *
''* Description: User interface functions                           *
''*      Author: Marko Lukat (c) 2013                               *
''*******************************************************************
'' Revision History
''-----------------
''   Release: 2013/10/01 - Version: 0.1.pal.1 - by Marko Lukat 
''     (original) initial version (640x480@60Hz timing, %11 sync locked)                              
''   Release: 2013/10/11 - Version: 0.2.pal.1 - by PropGuy2
''        (1) changed filename to: VGA_40x15_display.spin 
''        (2) update & rework: PUB out(c)
''        (3) added function:  PUB textXY(cx, cy, tx) 
''        (4) added function:  PUB gotoXY(cx, cy)  
''        (5) added function:  PUB Integer (charS) 
''        (6) delete function: PRI tr(c, addr, limit, response) 
''-----------------
''
CON
  columns  = driver#res_x / 16
  rows     = driver#res_y / 32
  bcnt     = columns * rows

  rows_raw = (driver#res_y + 32 - 1) / 32
  bcnt_raw = columns * rows_raw

  #8, BS, TAB, LF, VT, FF, CR, ESC = 27

  MAX_LEN = 64                        '63 chars + zero terminator
  
VAR
  long  link[driver#res_m]            'Driver mailbox
  word  scrn[bcnt_raw / 2]            'Screen buffer (2n aligned)
  word  indx[bcnt_raw / 2]            'Colour buffer (2n aligned)

  byte  x, y, page, colour            'Cursor position, page mode, colour
  word  flag

  long  idx                           'Pointer into string
  byte  nstr[MAX_LEN]  

OBJ
  driver: "VGA_40x15_driver"
 
PUB null
'' This is not a top level object.

PUB init                              'initialize VGA driver
  out($FF )                                     ' clear screen  
  link{0} := @scrn{0}                           ' initial screen and
  link[2] := @indx{0}                           ' colour index buffer (default palette)
  return driver.init(-1, @link{0})              ' video driver and pixel generator
  
PUB setn(n, addr)                     'set PaletteColor(0,1) function
  link[n] := addr
  repeat
  while link[n]
' return

PUB putc(c)                           'output Character to VGA screen
  indx.byte[y * columns + x] := colour      ' update colour
  scrn.byte[y * columns + x] := c           ' update character
  if ++x == columns                                       
    x := newline                            ' update CR/LF
' return

PUB str(addr)                         'output String to VGA screen
  repeat strsize(addr)
    out(byte[addr++])          
' return

PUB out(c) | i, k                     'output Character to VGA screen
'' Output character
''     $01 = clear screen
''     $02 = home x,y =0
''     $03 = page mode(s) 
''     $04 = page mode(S) 
''     $08 = backspace(x-1)
''     $09 = tab (8 spaces per)
''     $0A = set X position (X follows)
''     $0B = set Y position (Y follows)
''     $0C = set color  (color follows)
''     $0D = return ('newLine'function)
''     $FF = clear screen 
  case flag                 'check Flag index
    $00: case c             'check(char$) index
           $01:             'Clear screen 
              wordfill(@indx{0}, colour * $0101, constant(bcnt_raw / 2))             
              wordfill(@scrn{0}, $2020, constant(bcnt_raw / 2))
             'col := row := 0   'set Screen Clear w/color
              x   :=  y  := 0   'set XY cursor Home

           $FF:             'Clear screen
              wordfill(@indx{0}, colour * $0101, constant(bcnt_raw / 2))
              wordfill(@scrn{0}, $2020, constant(bcnt_raw / 2))
             'col := row := 0   'set Screen Clear w/color
              x   :=  y  := 0   'set XY cursor Home
                          
           $02:             'set Home
             'col := row := 0
              x   :=  y  := 0   'set XY cursor Home 
           $03:
              c := "s"      'set "s" Page = scroll mode
              page := c & $20       'page mode (ESC+s) (ESC+S) 
           $04:
              c := "S"      'set "S" Page = scroll mode
              page := c & $20       'page mode (ESC+s) (ESC+S)                        
           $08:             'set BackSpace
              x := (x - 1) #> 0                 
           $09:             'set Tab(8sp)
              repeat 8 - (x & 7)                    
                 putc(" ") 
              
           $0A:             'flag (X) cursor
              flag := c     'set Flag index
                   return   'and Return (for X following)
           $0B:             'flag (Y) cursor
              flag := c     'set Flag index
                   return   'and Return (for Y following)
           $0C:             'flag new Colour 
              flag := c     'set Flag index
                   return   'and Return (Color following)              
           $0D:             'set'NewLine' function
              x := newline  'set CR /LF scrolling
              
           other:
              putc(c)       'set Print Character       
    $0A:                    'get Flag index
        x := c -1           'set new (X) cursor
    $0B:                    'get Flag index 
        y := c -1           'set new (Y) cursor
    $0C:                    'get Flag index 
        colour := c         'set new Colour
  flag := 0                 'reset Flag index =0
 'return 

PUB textXY(cx, cy, tx)                'Text function @(X,Y)
  x := cx -1
  y := cy -1   
  str(tx)
' return  

PUB gotoXY(cx, cy)                    'Cursor function @(X,Y)
  x := cx -1
  y := cy -1
' return   

PUB Integer (charS)                   'Convert CharString to Number
    repeat (strsize(charS) <# 10)     'result = Integer
      result := ((result * 10) + (byte[charS++] & $F))

PRI newline                           'NewLine function
  if ++y == rows                      'with PageScrolling
    if page                           'page/scroll = TRUE
      y := 0
      return
    y--

    wordmove(@indx{0}, @indx.byte[columns], constant((bcnt_raw - columns) / 2))
    wordmove(@scrn{0}, @scrn.byte[columns], constant((bcnt_raw - columns) / 2))
    if rows_raw == rows
      wordfill(@indx.byte[constant(bcnt_raw - columns)], colour * $0101, constant(columns / 2))
      wordfill(@scrn.byte[constant(bcnt_raw - columns)], $2020, constant(columns / 2))
' return

DAT